# -*- coding: utf-8 -*-

##############################################################################
##
# This file is part of Taurus, a Tango User Interface Library
##
# http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

'''Utility functions to convert between tango and Taurus types'''

import PyTango

from taurus.core.units import Quantity, UR
from pint import UndefinedUnitError
from taurus.core.taurusbasetypes import (AttrQuality, DisplayLevel,
                                         TaurusAttrValue, DataType, DataFormat)


__NO_STR_VALUE = PyTango.constants.AlrmValueNotSpec, PyTango.constants.StatusNotSet

FROM_TANGO_TO_TAURUS_DFORMAT = {PyTango.AttrDataFormat.SCALAR: DataFormat._0D,
                                PyTango.AttrDataFormat.SPECTRUM: DataFormat._1D,
                                PyTango.AttrDataFormat.IMAGE: DataFormat._2D}

FROM_TANGO_TO_TAURUS_TYPE = {PyTango.CmdArgType.DevVoid: None,
                             PyTango.CmdArgType.DevBoolean: DataType.Boolean,
                             PyTango.CmdArgType.DevShort: DataType.Integer,
                             PyTango.CmdArgType.DevLong: DataType.Integer,
                             PyTango.CmdArgType.DevFloat: DataType.Float,
                             PyTango.CmdArgType.DevDouble: DataType.Float,
                             PyTango.CmdArgType.DevUShort: DataType.Integer,
                             PyTango.CmdArgType.DevULong: DataType.Integer,
                             PyTango.CmdArgType.DevString: DataType.String,
                             PyTango.CmdArgType.DevVarCharArray: DataType.Bytes,
                             PyTango.CmdArgType.DevVarShortArray: DataType.Integer,
                             PyTango.CmdArgType.DevVarLongArray: DataType.Integer,
                             PyTango.CmdArgType.DevVarFloatArray: DataType.Float,
                             PyTango.CmdArgType.DevVarDoubleArray: DataType.Float,
                             PyTango.CmdArgType.DevVarUShortArray: DataType.Integer,
                             PyTango.CmdArgType.DevVarULongArray: DataType.Integer,
                             PyTango.CmdArgType.DevVarStringArray: DataType.String,
                             PyTango.CmdArgType.DevVarLongStringArray: DataType.Object,
                             PyTango.CmdArgType.DevVarDoubleStringArray: DataType.Object,
                             PyTango.CmdArgType.DevState: DataType.DevState,
                             PyTango.CmdArgType.ConstDevString: DataType.String,
                             PyTango.CmdArgType.DevVarBooleanArray: DataType.Boolean,
                             PyTango.CmdArgType.DevUChar: DataType.Bytes,
                             PyTango.CmdArgType.DevLong64: DataType.Integer,
                             PyTango.CmdArgType.DevULong64: DataType.Integer,
                             PyTango.CmdArgType.DevVarLong64Array: DataType.Integer,
                             PyTango.CmdArgType.DevVarULong64Array: DataType.Integer,
                             PyTango.CmdArgType.DevInt: DataType.Integer,
                             PyTango.CmdArgType.DevEncoded: DataType.DevEncoded,
                             }

if hasattr(PyTango, "str_2_obj"):
    str_2_obj = PyTango.str_2_obj
else:

    # Old PyTango
    import PyTango.utils

    bool_ = lambda value_str: value_str.lower() == "true"

    def str_2_obj(obj_str, tg_type=None):
        f = str
        if PyTango.utils.is_scalar_type(tg_type):
            if PyTango.utils.is_numerical_type(tg_type):
                if obj_str in __NO_STR_VALUE:
                    return None

            if PyTango.utils.is_int_type(tg_type):
                f = int
            elif PyTango.utils.is_float_type(tg_type):
                f = float
            elif PyTango.utils.is_bool_type(tg_type):
                f = bool_
        return f(obj_str)


def get_quantity(value, units=None, fmt=None):
    if value is None:
        return None
    res = Quantity(value, units=units)
    if fmt is not None:
        res.default_format = fmt + res.default_format
    return res


def quantity_from_tango_str(value_str, dtype=None, units=None, fmt=None,
                            ignore_exception=True):
    try:
        return get_quantity(str_2_obj(value_str, dtype), units=units, fmt=fmt)
    except ValueError:
        if not ignore_exception:
            raise
        return None


def unit_from_tango(unit):
    from taurus import deprecated
    deprecated(dep='unit_from_tango', rel='4.0.4', alt="pint's parse_units")

    if unit == PyTango.constants.UnitNotSpec or unit == "No unit":
        unit = None
    try:
        return UR.parse_units(unit)
    except (UndefinedUnitError, UnicodeDecodeError):
        # TODO: Maybe we could dynamically register the unit in the UR
        from taurus import warning
        warning('Unknown unit "%s" (will be treated as unitless)', unit)
        return UR.parse_units(None)


def ndim_from_tango(data_format):
    return int(data_format)


def data_format_from_tango(data_format):
    return FROM_TANGO_TO_TAURUS_DFORMAT[data_format]


def data_type_from_tango(data_type):
    return FROM_TANGO_TO_TAURUS_TYPE[data_type]


def display_level_from_tango(disp_level):
    return DisplayLevel(disp_level)


def quality_from_tango(quality):
    return AttrQuality(int(quality))


__NULL_DESC = PyTango.constants.DescNotSet, PyTango.constants.DescNotSpec


def description_from_tango(desc):
    if desc in __NULL_DESC:
        desc = ''
    return desc

__S_TYPES = (PyTango.CmdArgType.DevString,
             PyTango.CmdArgType.DevVarStringArray,
             PyTango.CmdArgType.DevEncoded,)


def standard_display_format_from_tango(dtype, fmt):
    if fmt == 'Not specified':
        return '!s'

    # %6.2f is the default value that Tango sets when the format is
    # unassigned in tango < 8. This is only good for float types! So for other
    # types I am changing this value.
    if fmt == '%6.2f':
        if PyTango.is_float_type(dtype, inc_array=True):
            pass
        elif PyTango.is_int_type(dtype, inc_array=True):
            fmt = '%d'
        elif dtype in __S_TYPES:
            fmt = '%s'
    return fmt


def display_format_from_tango(dtype, fmt):
    fmt = standard_display_format_from_tango(dtype, fmt)
    return fmt.replace('%s', '!s').replace('%r', '!r').replace('%', '')





#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module defines the test suite for the whole Taurus package
Usage::

  from taurus.test import testsuite
  testsuite.run()

"""
from __future__ import print_function

import os
import sys
import re
import unittest
import taurus
from taurus.external.qt import PYQT4

__docformat__ = 'restructuredtext'

PY3_EXCLUDED = (
    'unittest.loader._FailedTest.taurus.qt.qtgui.qwt5',
    'unittest.loader._FailedTest.taurus.qt.qtgui.extra_sardana',
    'unittest.loader._FailedTest.taurus.qt.qtgui.extra_pool',
    'unittest.loader._FailedTest.taurus.qt.qtgui.extra_macroexecutor'
)

ONLY_PYQT4 = (
    'unittest.loader._FailedTest.taurus.qt.qtgui.qwt5',
)

def _filter_suite(suite, exclude_pattern, ret=None):
    """removes TestCases from a suite based on regexp matching on the Test id"""
    if ret is None:
        ret = unittest.TestSuite()
    for e in suite:
        if isinstance(e, unittest.TestCase):

            if e.__module__ == 'unittest.case':
                continue

            if sys.version_info.major > 2 and e.id() in PY3_EXCLUDED:
                print("Excluded %s" % e.id())
                continue

            if not PYQT4 and e.id() in ONLY_PYQT4:
                print("Excluded %s" % e.id())
                continue
            
            if re.match(exclude_pattern, e.id()):
                print("Excluded %s" % e.id())
                continue
            ret.addTest(e)
        else:
            _filter_suite(e, exclude_pattern, ret=ret)
    return ret


def get_taurus_suite(exclude_pattern='(?!)'):
    """discover all tests in taurus, except those matching `exclude_pattern`"""
    loader = unittest.defaultTestLoader
    start_dir = os.path.dirname(taurus.__file__)
    suite = loader.discover(start_dir, top_level_dir=os.path.dirname(start_dir))
    return _filter_suite(suite, exclude_pattern)


def run(disableLogger=True, exclude_pattern='(?!)'):
    """Runs tests for the taurus package"""
    # disable logging messages
    if disableLogger:
        taurus.disableLogOutput()
    # discover tests within the taurus/lib directory
    suite = get_taurus_suite(exclude_pattern=exclude_pattern)
    # use the basic text test runner that outputs to sys.stderr
    runner = unittest.TextTestRunner(descriptions=True, verbosity=2)
    # run the test suite
    return runner.run(suite)


def main():
    import taurus.test.skip
    import argparse
    from taurus import Release
    parser = argparse.ArgumentParser(description='Main test suite for Taurus')
    parser.add_argument('--skip-gui-tests', dest='skip_gui',
                        action='store_true', default=False,
                        help='Do not perform tests requiring GUI')
    # TODO: Define the default exclude patterns as a tauruscustomsettings
    # variable.
    help = """regexp pattern matching test ids to be excluded.
    (e.g. 'taurus\.core\..*' would exclude taurus.core tests)
    """
    parser.add_argument('-e', '--exclude-pattern',
                        dest='exclude_pattern',
                        default='(?!)',
                        help=help)
    parser.add_argument('--version', action='store_true', default=False,
                        help="show program's version number and exit")
    args = parser.parse_args()

    if args.version:
        print(Release.version)
        sys.exit(0)

    if args.skip_gui:
        import taurus.test.skip
        taurus.test.skip.GUI_TESTS_ENABLED = False
    if not taurus.test.skip.GUI_TESTS_ENABLED:
        exclude_pattern = '(taurus\.qt\..*)|(%s)' % args.exclude_pattern
    else:
        exclude_pattern = args.exclude_pattern

    ret = run(exclude_pattern=exclude_pattern)

    # calculate exit code (0 if OK and 1 otherwise)
    if ret.wasSuccessful():
        exit_code = 0
    else:
        exit_code = 1
    sys.exit(exit_code)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides taurus-specific functions related to icons"""

__all__ = [
    'sanitizePrefix',
    'registerPathFiles',
    'registerTheme',
    'getCachedPixmap',
    'getStandardIcon',
    'getElementTypeToolTip',
    'getElementTypeSize',
    'getElementTypeIcon',
    'getElementTypeIconName',
    'getElementTypePixmap',
    'getDevStateToolTip',
    'getDevStateIcon',
    'getDevStatePixmap',
    'REGISTERED_PREFIXES',
]

__docformat__ = 'restructuredtext'

import os
import re
import sys
import json

from taurus.external.qt import Qt

from taurus.core.taurusbasetypes import TaurusElementType, TaurusDevState
from taurus.core.util.log import Logger

__LOGGER = Logger(__name__)

__INITIALIZED = False

# Default width, height and QSize constants
__DW = 70
__DH = 24

__DQS = Qt.QSize(__DW, __DH)
__1DQS = __DQS
__2DQS = Qt.QSize(2 * __DW, __DH)
__3DQS = Qt.QSize(3 * __DW, __DH)

# Indexes for the map below
__IDX_ELEM_TYPE_ICON, __IDX_ELEM_TYPE_SIZE, __IDX_ELEM_TYPE_TOOLTIP = list(range(3))

# New default role map
# Elements are: icon theme, preferred size, description/tooltip
_ELEM_TYPE_MAP = {
    TaurusElementType.Name: ("folder", __3DQS, None),
    TaurusElementType.Device: ("applications-system", Qt.QSize(210, __DH),
                               "Tango device name"),
    TaurusElementType.DeviceAlias: ("applications-system", Qt.QSize(140, __DH),
                                    "Tango device alias"),
    TaurusElementType.Domain: ("folder", Qt.QSize(80, __DH),
                               "Tango device domain"),
    TaurusElementType.Family: ("folder", Qt.QSize(80, __DH),
                               "Tango device family"),
    TaurusElementType.Member: ("applications-system", Qt.QSize(80, __DH),
                               "Tango device member"),
    TaurusElementType.Server: ("application-x-executable", Qt.QSize(190, __DH),
                               "Tango server"),
    TaurusElementType.ServerName: ("application-x-executable",
                                   Qt.QSize(80, __DH), "Tango server name"),
    TaurusElementType.ServerInstance: ("application-x-executable",
                                       Qt.QSize(80, __DH),
                                       "Tango server instance"),
    TaurusElementType.DeviceClass: ("text-x-script", Qt.QSize(140, __DH),
                                    "Tango class name"),
    TaurusElementType.Exported: ("start-here", Qt.QSize(60, __DH),
                                 "Alive/not alive"),
    TaurusElementType.Host: ("network-server", Qt.QSize(100, __DH),
                             "Host machine were last ran"),
    TaurusElementType.Attribute: (":/actions/format-text-bold.svg",
                                  Qt.QSize(100, __DH), "Attribute name"),
}

# Indexes for the map below
__IDX_STATE_ICON, __IDX_STATE_TOOLTIP = list(range(2))

_STATE_MAP = {
    TaurusDevState.Ready: ("status:available.svg", "Element ready"),
    TaurusDevState.NotReady: ("status:not-available.svg", "Element not ready"),
    TaurusDevState.Undefined: ("status:not-known.svg",
                               "Element state undefined")
}

# set of registered prefixes (updated by registerPathFiles() )
REGISTERED_PREFIXES = set()


def sanitizePrefix(prefix):
    """strips any leading '/' and substitutes non alphanumeric characters by '_'
    """
    prefix = prefix.lstrip('/')
    return re.sub('[^0-9a-zA-Z]+', '_', prefix)


def registerPathFiles(pathfilenames):
    """
    Use given .path files to update Qt's search path
    Each path file contains a json-encoded list of (prefix,path) tuples.
    This function will call Qt.QDir.addSearchPath with each of the tuples
    from the path files (prefix values will be sanitized first, and relative
    path values will be made relative to the dir containing the .path file)

    :param pathfilenames: (list<str>) list of .path file names
    """
    for filename in pathfilenames:
        try:
            pathmap = json.load(open(filename))
        except Exception as e:
            __LOGGER.error('Error registering "%s": %r', filename, e)
            pathmap = []

        base_dir = os.path.abspath(os.path.dirname(filename))
        for prefix, path in pathmap:
            prefix = sanitizePrefix(prefix)
            path = os.path.join(base_dir, path)  # no effect if path is absolute
            Qt.QDir.addSearchPath(prefix, path)
            REGISTERED_PREFIXES.add(prefix)


def registerTheme(name='Tango', path='', force=False):
    """Use bundled them if OS does not define a theme (non-X11 systems)

    :param name: (str) icon theme name (default=Tango)
    :param path: (str) path to dir containing the theme (absolute or relative
                 to dir of taurus.qt.qtgui.icon). Default = ''
    :param force: (bool) Force to set path even if a theme is already set
    """
    if force or not sys.platform.startswith('linux'):
        base_dir = os.path.abspath(os.path.dirname(__file__))
        path = os.path.join(base_dir, path)
        Qt.QIcon.setThemeSearchPaths([path])
        Qt.QIcon.setThemeName(name)
        __LOGGER.info('Setting %s icon theme (from %s)', name, path)


def getCachedPixmap(key, size=None):
    """Returns a PyQt4.QtGui.QPixmap object for the given key and size.
    The key argument supports QDir's searchPath prefixes (see
    :meth:`QDir.setSearchPaths`).

    :param key: (str) the pixmap key., e.g.: 'status:folder-open.svg'
    :param size: (int) the pixmap size in pixels (will get a square pixmap).
                If None is passed it will return the original size

    :return: (PyQt4.QtGui.QPixmap)"""

    name = key
    if size is not None:
        key += "_%sx%s" % (size, size)
    pm = Qt.QPixmapCache.find(key)
    if pm is None:
        pm = Qt.QPixmap(name)
        if size is not None:
            pm = pm.scaled(size, size, Qt.Qt.KeepAspectRatio,
                           Qt.Qt.SmoothTransformation)
        Qt.QPixmapCache.insert(key, pm)
    return Qt.QPixmap(pm)


def getStandardIcon(key, widget=None):
    """Returns a PyQt4.QtGui.QIcon object for the given key. Key should be a
    QStyle.StandardPixmap enumeration member. The widget argument is optional
    and can also be used to aid the determination of the icon.

    :param key: (QStyle.StandardPixmap) a standard pixmap which can follow some
                existing GUI style or guideline
    :param widget: (Qt.QWidget) the widget argument (optional) can also be used
                   to aid the determination of the icon.

    :return: (PyQt4.QtGui.QIcon)"""
    styleOption = None
    if widget is not None:
        styleOption = Qt.QStyleOption()
        styleOption.initFrom(widget)
    style = Qt.QApplication.instance().style()
    return style.standardIcon(key, styleOption, widget)


def getElementTypeToolTip(elemType):
    data = _ELEM_TYPE_MAP.get(elemType)
    if data is None:
        return
    return data[__IDX_ELEM_TYPE_TOOLTIP]


def getElementTypeSize(elemType):
    data = _ELEM_TYPE_MAP.get(elemType)
    if data is None:
        return
    return data[__IDX_ELEM_TYPE_SIZE]


def getElementTypeIconName(elemType):
    """Gets an icon name string for the given
    :class:`taurus.core.taurusbasetypes.TaurusElementType`.

    If an icon name cannot be found for elemType, None is returned.

    :param elemType: (TaurusElementType) the taurus element type

    :return: (str) a string representing the icon name for the given
             :class:`taurus.core.taurusbasetypes.TaurusElementType`"""
    if elemType is None:
        return
    data = _ELEM_TYPE_MAP.get(elemType)
    if data is None:
        return
    return data[__IDX_ELEM_TYPE_ICON]


def getElementTypeIcon(elemType, fallback=None):
    """Gets a PyQt4.QtGui.QIcon object for the given
    :class:`taurus.core.taurusbasetypes.TaurusElementType`.

    If an icon cannot be found for the given TaurusElementType,
    fallback is returned.

    :param elemType: (TaurusElementType) the taurus element type
    :param fallback: (PyQt4.QtGui.QIcon) the fallback icon. Default is None.

    :return: (PyQt4.QtGui.QIcon)"""

    themeIconName = getElementTypeIconName(elemType)
    icon = Qt.QIcon.fromTheme(themeIconName)
    if icon.isNull() and fallback is not None:
        icon = fallback
    return icon


def getElementTypePixmap(elemType, size=None):
    """Gets a PyQt4.QtGui.QPixmap object for the given
    :class:`taurus.core.taurusbasetypes.TaurusElementType`.

    :param elemType: (TaurusElementType) the taurus element type
    :param size: (int) the pixmap size in pixels (will get a square pixmap).
                 Default is None meaning it will return the original size.

    :return: (PyQt4.QtGui.QPixmap or None)"""

    if elemType is None:
        return
    data = _ELEM_TYPE_MAP.get(elemType)
    if data is None:
        return
    themeKey = data[__IDX_ELEM_TYPE_ICON]
    return Qt.QIcon.fromTheme(themeKey, size).pixmap(size, size)


def getDevStateToolTip(state):
    data = _STATE_MAP.get(state)
    if data is None:
        return
    return data[__IDX_STATE_TOOLTIP]


def getDevStateIcon(state, fallback=None):
    """Gets a PyQt4.QtGui.QIcon object for the given
    :class:`taurus.core.taurusbasetypes.TaurusDevState`.

    If an icon cannot be found for the given state, fallback is returned.

    :param state: (TaurusDevState) the taurus device state
    :param fallback: (PyQt4.QtGui.QIcon) the fallback icon. Default is None.

    :return: (PyQt4.QtGui.QIcon or None)"""
    if state is None:
        return
    data = _STATE_MAP.get(state)
    if data is None:
        return
    name = data[__IDX_STATE_ICON]

    icon = Qt.QIcon(name)
    if icon.isNull() and fallback is not None:
        icon = fallback
    return icon


def getDevStatePixmap(state, size=None):
    """Gets a PyQt4.QtGui.QPixmap object for the given
    :class:`taurus.core.taurusbasetypes.TaurusDevState`.

    :param state: (TaurusDevState) the taurus software device state
    :param size: (int) the pixmap size in pixels (will get a square pixmap).
                 Default is None meaning it will return the original size.

    :return: (PyQt4.QtGui.QPixmap or None)
    """
    if state is None:
        return None
    data = _STATE_MAP.get(state)
    if data is None:
        return None
    name = data[__IDX_STATE_ICON]
    return getCachedPixmap(name, size)

if __name__ == '__main__':
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication()

    icons = [
        # null because of non-existent path
        Qt.QIcon('NONEXISTENT/PATH'),
        # null because of resource-style key (and no resources defined)
        Qt.QIcon(':/actions/edit-cut.svg'),
        # abs path
        Qt.QIcon(os.path.abspath('Tango/scalable/actions/edit-cut.svg')),
        # rel path
        Qt.QIcon('Tango/scalable/actions/edit-cut.svg'),
        # New-style (using prefix)
        Qt.QIcon('actions:edit-cut.svg'),
        Qt.QIcon("apps:preferences-system-session.svg"),
        Qt.QIcon("designer:devs_tree.png"),
        Qt.QIcon("actions:process-stop.svg"),  # from tango-icons/actions
        Qt.QIcon("actions:add.svg"), # from rrze-icons/actions
        Qt.QIcon("actions:stop.svg"), # from extra-icons/actions
        Qt.QIcon("logos:taurus.svg"), # uses the "general",
        Qt.QIcon.fromTheme('computer')
        ]

    w = Qt.QWidget()
    l = Qt.QVBoxLayout()
    w.setLayout(l)
    for icon in icons:
        button = Qt.QPushButton(icon, 'kk')
        l.addWidget(button)

    w.show()

    sys.exit(app.exec_())



"""Provide a Signal class for non-QObject objects"""
from __future__ import print_function

from builtins import object

from taurus.external.qt import Qt
from threading import Lock
from weakref import WeakKeyDictionary

__all__ = ['baseSignal']


def baseSignal(name, *args):
    """Signal class for non-Qobject objects.

    :param name: signal name (unlike pyqtSignal, name has
                 to be specified explicitely)
    :param args: arguments passed to the pyqtSignal
    """
    main_lock = Lock()
    lock_dict = WeakKeyDictionary()
    classname = name.capitalize() + "Signaller"
    attrs = {name: Qt.pyqtSignal(*args, name=name)}
    signaller_type = type(classname, (Qt.QObject,), attrs)

    def get_lock(self):
        with main_lock:
            if self not in lock_dict:
                lock_dict[self] = Lock()
        return lock_dict[self]

    def get_signaller(self):
        with get_lock(self):
            if not hasattr(self, '_signallers'):
                self._signallers = {}
            if (name, args) not in self._signallers:
                self._signallers[name, args] = signaller_type()
        return self._signallers[name, args]

    def get_signal(self):
        return getattr(get_signaller(self), name)

    doc = "Base signal {}".format(name)
    return property(get_signal, doc=doc)

if __name__ == '__main__':

    class Base(object):
        test = baseSignal('test', int)

    def print_func(arg):
        print(arg)

    base1, base2 = Base(), Base()
    base1.test.connect(print_func)
    base2.test.connect(print_func)
    base1.test.emit(1)
    base2.test.emit(2)



# -*- coding: utf-8 -*-

##############################################################################
##
# This file is part of Taurus, a Tango User Interface Library
##
# http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

__all__ = ["tangoFormatter"]

from taurus.core.units import Quantity

def tangoFormatter(dtype=None, **kwargs):
    """
    The tango formatter callable. Returns a format string based on
    the `format` Tango Attribute configuration (Display.Format in Tango DB)

    :param dtype: (type) type of the value object
    :param kwargs: other keyword arguments (ignored)

    :return: the string formatting
    """
    if dtype is Quantity:
        fmt = "{:~{bc.modelObj.format_spec}}"
    else:
        fmt = "{:{bc.modelObj.format_spec}}"
    return fmt



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a set of useful logging elements based on python's
:mod:`logging` system."""
from __future__ import print_function
from __future__ import absolute_import

from builtins import str
from builtins import object

import io
import os
import sys
import logging.handlers
import weakref
import warnings
import traceback
import inspect
import threading
import functools

from .object import Object
from .wrap import wraps
from .excepthook import BaseExceptHook

# ------------------------------------------------------------------------------
# TODO: substitute this ugly hack (below) by a more general mechanism
from collections import defaultdict


__all__ = ["LogIt", "TraceIt", "DebugIt", "InfoIt", "WarnIt", "ErrorIt",
           "CriticalIt", "MemoryLogHandler", "LogExceptHook", "Logger",
           "LogFilter",
           "_log", "trace", "debug", "info", "warning", "error", "fatal",
           "critical", "deprecated", "deprecation_decorator",
           "taurus4_deprecation"]

__docformat__ = "restructuredtext"


class _DeprecationCounter(defaultdict):

    def __init__(self):
        defaultdict.__init__(self, int)

    def getTotal(self):
        c = 0
        for v in self.values():
            c += v
        return c

    def pretty(self):
        from operator import itemgetter
        sorted_items = sorted(self.items(), key=itemgetter(1), reverse=True)
        ret = '\n'.join(['\t%d * "%s"' % (v, k) for k, v in sorted_items])
        return "< Deprecation Counts (%d):\n%s >" % (self.getTotal(), ret)

_DEPRECATION_COUNT = _DeprecationCounter()
# ------------------------------------------------------------------------------

TRACE = 5
logging.addLevelName(TRACE, "TRACE")

#
# _srcfile is used when walking the stack to check when we've got the first
# caller stack frame.
#
if hasattr(sys, 'frozen'):  # support for py2exe
    _srcfile = "logging%s__init__%s" % (os.sep, __file__[-4:])
elif __file__[-4:].lower() in ['.pyc', '.pyo']:
    _srcfile = __file__[:-4] + '.py'
else:
    _srcfile = __file__
_srcfile = os.path.normcase(_srcfile)

# next bit filched from 1.5.2's inspect.py


def currentframe():
    """Return the frame object for the caller's stack frame."""
    try:
        raise Exception
    except:
        return sys.exc_info()[2].tb_frame.f_back

if hasattr(sys, '_getframe'):
    currentframe = lambda: sys._getframe(3)
# done filching


class LogIt(object):
    """A function designed to be a decorator of any method of a Logger subclass.
    The idea is to log the entrance and exit of any decorated method of a Logger
    subclass.
    Example::

        from taurus.core.util.log import Logger, LogIt

        class Example(Logger):

            @LogIt(Logger.Debug)
            def go(self):
                print "Hello world "

    This will generate two log messages of Debug level, one before the function
    go is called and another when go finishes. Example output::

        MainThread     DEBUG    2010-11-15 15:36:11,440 Example: -> go
        Hello world of mine
        MainThread     DEBUG    2010-11-15 15:36:11,441 Example: <- go

    This decorator can receive two optional arguments **showargs** and **showret**
    which are set to False by default. Enabling them will had verbose infomation
    about the parameters and return value. The following example::

        from taurus.core.uti.log import Logger, LogIt

        class Example(Logger):

            @LogIt(Logger.Debug, showargs=True, showret=True)
            def go(self, msg):
                msg = "Hello world",msg
                print msg
                return msg

    would generate an output like::

        MainThread     DEBUG    2010-11-15 15:42:02,353 Example: -> go('of mine',)
        Hello world of mine
        MainThread     DEBUG    2010-11-15 15:42:02,353 Example: <- go = Hello world of mine

    .. note::
        it may happen that in these examples that the output of the method
        appears before or after the log messages. This is because log
        messages are, by default, written to the *stardard error* while the print
        message inside the go method outputs to the *standard ouput*. On many
        systems these two targets are not synchronized.
    """

    def __init__(self, level=logging.DEBUG, showargs=False, showret=False, col_limit=0):
        self._level = level
        self._showargs = showargs
        self._showret = showret
        self._col_limit = col_limit

    def __call__(self, f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            f_self = args[0]
            if f_self.log_level > self._level:
                return f(*args, **kwargs)

            has_log = hasattr(f_self, "log")
            fname = f.__name__
            log_obj = f_self
            if not has_log:
                log_obj = logging.getLogger()
                try:
                    fname = "%s.%s" % (f_self.__class__.__name__, fname)
                except:
                    pass
            in_msg = "-> %s" % fname
            if self._showargs:
                if len(args) > 1:
                    in_msg += str(args[1:])
                if len(kwargs):
                    in_msg += str(kwargs)
            if self._col_limit and len(in_msg) > self._col_limit:
                in_msg = "%s [...]" % in_msg[:self._col_limit - 6]
            log_obj.log(self._level, in_msg)
            out_msg = "<-"
            try:
                ret = f(*args, **kwargs)
            except Exception as e:
                exc_info = sys.exc_info()
                out_msg += " (with %s) %s" % (e.__class__.__name__, fname)
                log_obj.log(self._level, out_msg, exc_info=exc_info)
                raise
            out_msg += " %s" % fname
            if not ret is None and self._showret:
                out_msg += " = %s" % str(ret)
            if self._col_limit and len(out_msg) > self._col_limit:
                out_msg = "%s [...]" % out_msg[:self._col_limit - 6]
            log_obj.log(self._level, out_msg)
            return ret
        return wrapper


class TraceIt(LogIt):
    """Specialization of LogIt for trace level messages.
    Example::

        from taurus.core.util.log import Logger, TraceIt
        class Example(Logger):

            @TraceIt()
            def go(self):
                print "Hello world"

    .. seealso:: :class:`LogIt`"""

    def __init__(self, showargs=False, showret=False):
        LogIt.__init__(self, level=TRACE, showargs=showargs, showret=showret)


class DebugIt(LogIt):
    """Specialization of LogIt for debug level messages.
    Example::

        from taurus.core.util.log import Logger, DebugIt
        class Example(Logger):

            @DebugIt()
            def go(self):
                print "Hello world"

    .. seealso:: :class:`LogIt`"""

    def __init__(self, showargs=False, showret=False):
        LogIt.__init__(self, level=logging.DEBUG,
                       showargs=showargs, showret=showret)


class InfoIt(LogIt):
    """Specialization of LogIt for info level messages.
    Example::

        from taurus.core.util.log import Logger, InfoIt
        class Example(Logger):

            @InfoIt()
            def go(self):
                print "Hello world"

    .. seealso:: :class:`LogIt`"""

    def __init__(self, showargs=False, showret=False):
        LogIt.__init__(self, level=logging.INFO,
                       showargs=showargs, showret=showret)


class WarnIt(LogIt):
    """Specialization of LogIt for warn level messages.
    Example::

        from taurus.core.util.log import Logger, WarnIt
        class Example(Logger):

            @WarnIt()
            def go(self):
                print "Hello world"

    .. seealso:: :class:`LogIt`"""

    def __init__(self, showargs=False, showret=False):
        LogIt.__init__(self, level=logging.WARN,
                       showargs=showargs, showret=showret)


class ErrorIt(LogIt):
    """Specialization of LogIt for error level messages.
    Example::

        from taurus.core.util.log import Logger, ErrorIt
        class Example(Logger):

            @ErrorIt()
            def go(self):
                print "Hello world"

    .. seealso:: :class:`LogIt`"""

    def __init__(self, showargs=False, showret=False):
        LogIt.__init__(self, level=logging.ERROR,
                       showargs=showargs, showret=showret)


class CriticalIt(LogIt):
    """Specialization of LogIt for critical level messages.
    Example::

        from taurus.core.util.log import Logger, CriticalIt
        class Example(Logger):

            @CriticalIt()
            def go(self):
                print "Hello world"

    .. seealso:: :class:`LogIt`"""

    def __init__(self, showargs=False, showret=False):
        LogIt.__init__(self, level=logging.CRITICAL,
                       showargs=showargs, showret=showret)


class PrintIt(object):
    '''A decorator similar to TraceIt, DebugIt,... etc but which does not
    require the decorated class to inherit from Logger.
    It just uses print statements instead of logging. It is here just to be
    used as a replacement of those decorators if you cannot use them on a
    non-logger class.
    '''

    def __init__(self, showargs=False, showret=False):
        self._showargs = showargs
        self._showret = showret

    def __call__(self, f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            fname = f.__name__
            in_msg = "-> %s" % fname
            if self._showargs:
                if len(args) > 1:
                    in_msg += str(args[1:])
                if len(kwargs):
                    in_msg += str(kwargs)
            print()
            print(in_msg)
            out_msg = "<-"
            try:
                ret = f(*args, **kwargs)
            except Exception as e:
                out_msg += " (with %s) %s" % (e.__class__.__name__, fname)
                print(out_msg)
                raise
            out_msg += " %s" % fname
            if not ret is None and self._showret:
                out_msg += " = %s" % str(ret)
            print(out_msg)
            print()
            return ret
        return wrapper


class MemoryLogHandler(list, logging.handlers.BufferingHandler):
    """An experimental log handler that stores temporary records in memory.
       When flushed it passes the records to another handler"""

    def __init__(self, capacity=1000):
        list.__init__(self)
        logging.handlers.BufferingHandler.__init__(self, capacity=capacity)
        self._handler_list_changed = False

    def shouldFlush(self, record):
        """Determines if the given record should trigger the flush

           :param record: (logging.LogRecord) a log record
           :return: (bool) wheter or not the handler should be flushed
        """
        return (len(self.buffer) >= self.capacity) or \
            (record.levelno >= Logger.getLogLevel()) or \
            self._handler_list_changed

    def flush(self):
        """Flushes this handler"""
        for record in self.buffer:
            for handler in self:
                handler.handle(record)
        self.buffer = []

    def close(self):
        """Closes this handler"""
        self.flush()
        del self[:]
        logging.handlers.BufferingHandler.close(self)


class LogExceptHook(BaseExceptHook):
    """A callable class that acts as an excepthook that logs the exception in
    the python logging system.

    :param hook_to: callable excepthook that will be called at the end of
                    this hook handling [default: None]
    :type hook_to: callable
    :param name: logger name [default: None meaning use class name]
    :type name: str
    :param level: log level [default: logging.ERROR]
    :type level: int"""

    def __init__(self, hook_to=None, name=None, level=logging.ERROR):
        BaseExceptHook.__init__(self, hook_to=hook_to)
        name = name or self.__class__.__name__
        self._level = level
        self._log = Logger(name=name)

    def report(self, *exc_info):
        text = "".join(traceback.format_exception(*exc_info))
        if text[-1] == '\n':
            text = text[:-1]
        self._log.log(self._level, "Unhandled exception:\n%s", text)


class Logger(Object):
    """The taurus logger class. All taurus pertinent classes should inherit
    directly or indirectly from this class if they need taurus logging
    facilities."""

    #: Internal usage
    root_inited = False

    #: Internal usage
    root_init_lock = threading.Lock()

    #: Critical message level (constant)
    Critical = logging.CRITICAL

    #: Fatal message level (constant)
    Fatal = logging.FATAL

    #: Error message level (constant)
    Error = logging.ERROR

    #: Warning message level (constant)
    Warning = logging.WARNING

    #: Info message level (constant)
    Info = logging.INFO

    #: Debug message level (constant)
    Debug = logging.DEBUG

    #: Trace message level (constant)
    Trace = TRACE

    #: Default log level (constant)
    DftLogLevel = Info

    #: Default log message format (constant)
    DftLogMessageFormat = '%(threadName)-14s %(levelname)-8s %(asctime)s %(name)s: %(message)s'

    #: Default log format (constant)
    DftLogFormat = logging.Formatter(DftLogMessageFormat)

    #: Current global log level
    log_level = DftLogLevel

    #: Default log message format
    log_format = DftLogFormat

    #: the main stream handler
    stream_handler = None

    def __init__(self, name='', parent=None, format=None):
        """The Logger constructor

        :param name: (str) the logger name (default is empty string)
        :param parent: (Logger) the parent logger or None if no parent exists (default is None)
        :param format: (str) the log message format or None to use the default log format (default is None)
        """
        self.call__init__(Object)

        if format:
            self.log_format = format
        Logger.initRoot()

        if name is None or len(name) == 0:
            name = self.__class__.__name__
        self.log_name = name
        if parent is not None:
            self.log_full_name = '%s.%s' % (parent.log_full_name, name)
        else:
            self.log_full_name = name

        self.log_obj = self._getLogger(self.log_full_name)
        self.log_handlers = []

        self.log_parent = None
        self.log_children = {}
        if parent is not None:
            self.log_parent = weakref.ref(parent)
            parent.addChild(self)

    def cleanUp(self):
        """The cleanUp. Default implementation does nothing
           Overwrite when necessary"""
        pass

    @classmethod
    def initRoot(cls):
        """Class method to initialize the root logger. Do **NOT** call this
           method directly in your code
        """
        if cls.root_inited:
            return cls._getLogger()

        try:
            cls.root_init_lock.acquire()
            root_logger = cls._getLogger()
            logging.addLevelName(cls.Trace, "TRACE")
            cls.stream_handler = logging.StreamHandler(sys.__stderr__)
            cls.stream_handler.setFormatter(cls.log_format)
            root_logger.addHandler(cls.stream_handler)

            console_log_level = os.environ.get("TAURUSLOGLEVEL", None)
            if console_log_level is not None:
                console_log_level = console_log_level.capitalize()
                if hasattr(cls, console_log_level):
                    cls.log_level = getattr(cls, console_log_level)
            root_logger.setLevel(cls.log_level)
            Logger.root_inited = True
        finally:
            cls.root_init_lock.release()
        return root_logger

    @classmethod
    def addRootLogHandler(cls, h):
        """Adds a new handler to the root logger

           :param h: (logging.Handler) the new log handler
        """
        h.setFormatter(cls.getLogFormat())
        cls.initRoot().addHandler(h)

    @classmethod
    def removeRootLogHandler(cls, h):
        """Removes the given handler from the root logger

           :param h: (logging.Handler) the handler to be removed
        """
        cls.initRoot().removeHandler(h)

    @classmethod
    def enableLogOutput(cls):
        """Enables the :class:`logging.StreamHandler` which dumps log records,
           by default, to the stderr.
        """
        cls.initRoot().addHandler(cls.stream_handler)

    @classmethod
    def disableLogOutput(cls):
        """Disables the :class:`logging.StreamHandler` which dumps log records,
           by default, to the stderr.
        """
        cls.initRoot().removeHandler(cls.stream_handler)

    @classmethod
    def setLogLevel(cls, level):
        """sets the new log level (the root log level)

           :param level: (int) the new log level
        """
        cls.log_level = level
        cls.initRoot().setLevel(level)

    @classmethod
    def getLogLevel(cls):
        """Retuns the current log level (the root log level)

           :return: (int) a number representing the log level
        """
        return cls.log_level

    @classmethod
    def setLogFormat(cls, format):
        """sets the new log message format

           :param level: (str) the new log message format
        """
        cls.log_format = logging.Formatter(format)
        root_logger = cls.initRoot()
        for h in root_logger.handlers:
            h.setFormatter(cls.log_format)

    @classmethod
    def getLogFormat(cls):
        """Retuns the current log message format (the root log format)

           :return: (str) the log message format
        """
        return cls.log_format

    @classmethod
    def resetLogLevel(cls):
        """Resets the log level (the root log level)"""
        cls.setLogLevel(cls.DftLogLevel)

    @classmethod
    def resetLogFormat(cls):
        """Resets the log message format (the root log format)"""
        cls.setLogFormat(cls.DftLogFormat)

    @classmethod
    def addLevelName(cls, level_no, level_name):
        """Registers a new log level

           :param level_no: (int) the level number
           :param level_name: (str) the corresponding name
        """
        logging.addLevelName(level_no, level_name)
        level_name = level_name.capitalize()
        if not hasattr(cls, level_name):
            setattr(cls, level_name, level_no)

    @classmethod
    def getRootLog(cls):
        """Retuns the root logger

           :return: (logging.Logger) the root logger
        """
        return cls.initRoot()

    @staticmethod
    def _getLogger(name=None):
        orig_logger_class = logging.getLoggerClass()
        try:
            logging.setLoggerClass(logging.Logger)
            ret = logging.getLogger(name)
            return ret
        finally:
            logging.setLoggerClass(orig_logger_class)

    @classmethod
    def getLogger(cls, name=None):
        cls.initRoot()
        return cls._getLogger(name=name)

    def getLogObj(self):
        """Returns the log object for this object

           :return: (logging.Logger) the log object
        """
        return self.log_obj

    def getParent(self):
        """Returns the log parent for this object or None if no parent exists

           :return: (logging.Logger or None) the log parent for this object
        """
        if self.log_parent is None:
            return None
        return self.log_parent()

    def getChildren(self):
        """Returns the log children for this object

           :return: (sequence<logging.Logger) the list of log children
        """
        children = []
        for _, ref in self.log_children.items():
            child = ref()
            if child is not None:
                children.append(child)
        return children

    def addChild(self, child):
        """Adds a new logging child

           :param child: (logging.Logger) the new child
        """
        if not self.log_children.get(id(child)):
            self.log_children[id(child)] = weakref.ref(child)

    def addLogHandler(self, handler):
        """Registers a new handler in this object's logger

           :param handler: (logging.Handler) the new handler to be added
        """
        self.log_obj.addHandler(handler)
        self.log_handlers.append(handler)

    def removeLogHandler(self, handler):
        """Removes the given handler from this object's logger

           :param handler: (logging.Handler) the handler to be removed
        """
        self.log_obj.removeHandler(handler)
        self.log_handlers.remove(handler)

    def copyLogHandlers(self, other):
        """Copies the log handlers of other object to this object

           :param other: (object) object which contains 'log_handlers'
        """
        for handler in other.log_handlers:
            self.addLogHandler(handler)

    def trace(self, msg, *args, **kw):
        """Record a trace message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.log`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.log(self.Trace, msg, *args, **kw)

    def traceback(self, level=Trace, extended=True):
        """Log the usual traceback information, followed by a listing of all the
           local variables in each frame.

           :param level: (int) the log level assigned to the traceback record
           :param extended: (bool) if True, the log record message will have multiple lines

           :return: (str) The traceback string representation
        """
        out = traceback.format_exc()
        if extended:
            out += "\n"
            out += self._format_trace()

        self.log_obj.log(level, out)
        return out

    def stack(self, target=Trace):
        """Log the usual stack information, followed by a listing of all the
           local variables in each frame.

           :param target: (int) the log level assigned to the record

           :return: (str) The stack string representation
        """
        out = self._format_stack()
        self.log_obj.log(target, out)
        return out

    def _format_trace(self):
        return self._format_stack(inspect.trace)

    def _format_stack(self, stack_func=inspect.stack):
        line_count = 3
        stack = stack_func(line_count)
        out = ''
        for frame_record in stack:
            out += '\n\t' + 60 * '-'
            frame, filename, line, funcname, lines, index = frame_record
            #out += '\n\t    depth = %d' % frame[5]
            out += '\n\t filename = %s' % filename
            out += '\n\t function = %s' % funcname
            if lines is None:
                code = '<code could not be found>'
                out += '\n\t     line = [%d]: %s' % (line, code)
            else:
                lines, line_nb = [s.strip(' \n') for s in lines], len(lines)
                if line_nb >= 3:
                    out += '\n\t     line = [%d]: %s' % (line - 1, lines[0])
                    out += '\n\t  -> line = [%d]: %s' % (line, lines[1])
                    out += '\n\t     line = [%d]: %s' % (line + 1, lines[2])
                elif line_nb > 0:
                    out += '\n\t  -> line = [%d]: %s' % (line, lines[0])
            if frame:
                out += '\n\t   locals = '
                for k, v in frame.f_locals.items():
                    out += '\n\t\t%20s = ' % k
                    try:
                        cut = False
                        v = str(v)
                        i = v.find('\n')
                        if i == -1:
                            i = 80
                        else:
                            i = min(i, 80)
                            cut = True
                        if len(v) > 80:
                            cut = True
                        out += v[:i]
                        if cut:
                            out += '[...]'
                    except:
                        out += '<could not find suitable string representation>'
        return out

    def log(self, level, msg, *args, **kw):
        """Record a log message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.log`.

           :param level: (int) the record level
           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.log(level, msg, *args, **kw)

    def debug(self, msg, *args, **kw):
        """Record a debug message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.debug`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.debug(msg, *args, **kw)

    def info(self, msg, *args, **kw):
        """Record an info message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.info`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.info(msg, *args, **kw)

    def warning(self, msg, *args, **kw):
        """Record a warning message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.warning`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.warning(msg, *args, **kw)

    def deprecated(self, msg=None, dep=None, alt=None, rel=None, dbg_msg=None,
                   _callerinfo=None, **kw):
        """Record a deprecated warning message in this object's logger.
           If message is not passed, a estandard deprecation message is
           constructued using dep, alt, rel arguments.
           Also, an extra debug message can be recorded, followed by traceback
           info.

           :param msg: (str) the message to be recorded (if None passed, it will
                       be constructed using dep (and, optionally, alt and rel)
           :param dep: (str) name of deprecated feature (in case msg is None)
           :param alt: (str) name of alternative feature (in case msg is None)
           :param rel: (str) name of release from which the feature was
                       deprecated (in case msg is None)
           :param dbg_msg: (str) msg for debug (or None to log only the warning)
           :param _callerinfo: for internal use only. Do not use this argument.
           :param kw: any additional keyword arguments, are passed to
                     :meth:`logging.Logger.warning`
        """
        if msg is None:
            if dep is None:
                raise TypeError('deprecated takes either msg or dep argument')
            msg = '%s is deprecated' % dep
            if rel is not None:
                msg += ' since %s' % rel
            if alt is not None:
                msg += '. Use %s instead' % alt

        # count the number of calls (classified by msg)
        # TODO: substitute this ugly hack (below) by a more general mechanism
        _DEPRECATION_COUNT[msg] += 1
        # limit the output to 1 deprecation message of each type
        from taurus import tauruscustomsettings
        _MAX_DEPRECATIONS_LOGGED = getattr(tauruscustomsettings,
                                           '_MAX_DEPRECATIONS_LOGGED', None)
        if _MAX_DEPRECATIONS_LOGGED is not None:
            if _MAX_DEPRECATIONS_LOGGED < 0:
                self.stack(self.Warning)
                raise Exception(msg)
            if _DEPRECATION_COUNT[msg] > _MAX_DEPRECATIONS_LOGGED:
                return
        if _callerinfo is None:
            _callerinfo = self.log_obj.findCaller()
        filename, lineno = _callerinfo[:2]
        depr_msg = warnings.formatwarning(
            msg, DeprecationWarning, filename, lineno)
        self.log_obj.warning(depr_msg, **kw)
        if dbg_msg:
            self.debug(dbg_msg)
            self.stack()

    def error(self, msg, *args, **kw):
        """Record an error message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.error`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.error(msg, *args, **kw)

    def fatal(self, msg, *args, **kw):
        """Record a fatal message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.fatal`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.fatal(msg, *args, **kw)

    def critical(self, msg, *args, **kw):
        """Record a critical message in this object's logger. Accepted *args* and
           *kwargs* are the same as :meth:`logging.Logger.critical`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
           :param kw: list of keyword arguments
        """
        self.log_obj.critical(msg, *args, **kw)

    def exception(self, msg, *args):
        """Log a message with severity 'ERROR' on the root logger, with
           exception information.. Accepted *args* are the same as
           :meth:`logging.Logger.exception`.

           :param msg: (str) the message to be recorded
           :param args: list of arguments
        """
        self.log_obj.exception(msg, *args)

    def flushOutput(self):
        """Flushes the log output"""
        self.syncLog()

    def syncLog(self):
        """Synchronises the log output"""
        logger = self
        synced = []
        while logger is not None:
            for handler in logger.log_handlers:
                if handler in synced:
                    continue
                try:
                    sync = getattr(handler, 'sync')
                except:
                    continue
                sync()
                synced.append(handler)
            logger = logger.getParent()

    def getLogName(self):
        """Gets the log name for this object

           :return: (str) the log name
        """
        return self.log_name

    def getLogFullName(self):
        """Gets the full log name for this object

           :return: (str) the full log name
        """
        return self.log_full_name

    def changeLogName(self, name):
        """Change the log name for this object.

           :param name: (str) the new log name
        """
        self.log_name = name
        p = self.getParent()
        if p is not None:
            self.log_full_name = '%s.%s' % (p.log_full_name, name)
        else:
            self.log_full_name = name

        self.log_obj = logging.getLogger(self.log_full_name)
        for handler in self.log_handlers:
            self.log_obj.addHandler(handler)

        for child in self.getChildren():
            child.changeLogName(child.log_name)


class LogFilter(logging.Filter):
    """Experimental log filter"""

    def __init__(self, level):
        self.filter_level = level
        logging.Filter.__init__(self)

    def filter(self, record):
        ok = (record.levelno == self.filter_level)
        return ok


def __getrootlogger():
    return Logger.getLogger("TaurusRootLogger")

# cannot export log because upper package taurus.core.util imports this 'log'
# module and it would itself be overwritten by this log function


def _log(level, msg, *args, **kw):
    return __getrootlogger().log(level, msg, *args, **kw)


def trace(msg, *args, **kw):
    return _log(Logger.Trace, msg, *args, **kw)


def debug(msg, *args, **kw):
    return __getrootlogger().debug(msg, *args, **kw)


def info(msg, *args, **kw):
    return __getrootlogger().info(msg, *args, **kw)


def warning(msg, *args, **kw):
    return __getrootlogger().warning(msg, *args, **kw)


def error(msg, *args, **kw):
    return __getrootlogger().error(msg, *args, **kw)


def fatal(msg, *args, **kw):
    return __getrootlogger().fatal(msg, *args, **kw)


def critical(msg, *args, **kw):
    return __getrootlogger().critical(msg, *args, **kw)


def deprecated(*args, **kw):
    kw['_callerinfo'] = __getrootlogger().findCaller()
    return Logger("TaurusRootLogger").deprecated(*args, **kw)


def deprecation_decorator(func=None, alt=None, rel=None, dbg_msg=None):
    """decorator to mark methods as deprecated"""
    if func is None:
        return functools.partial(deprecation_decorator, alt=alt, rel=rel,
                                 dbg_msg=dbg_msg)

    def new_func(*args, **kwargs):
        deprecated(dep=func.__name__, alt=alt, rel=rel, dbg_msg=dbg_msg)
        return func(*args, **kwargs)

    doc = (func.__doc__ or '')
    doc += '\n\n.. deprecated:: %s\n' % (rel or '')
    if alt:
        doc += '   Use %s instead\n' % alt

    new_func.__name__ = func.__name__
    new_func.__doc__ = doc
    new_func.__dict__.update(func.__dict__)
    return new_func


taurus4_deprecation = functools.partial(deprecation_decorator, rel='4.0')


if __name__ == '__main__':

    @taurus4_deprecation(alt='bar')
    def foo(x):
        """Does this and that and also:

        - baz
        - zab
        """

    print(foo.__doc__)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""DataExportDlg.py: A Qt dialog for showing and exporting x-y Ascii data from
one or more curves"""

from __future__ import print_function

import os.path
from datetime import datetime

from taurus.external.qt import Qt, compat
from taurus.qt.qtgui.util.ui import UILoadable


__all__ = ["QDataExportDialog"]


@UILoadable
class QDataExportDialog(Qt.QDialog):
    """
    This creates a Qt dialog for showing and exporting x-y Ascii data from one or more curves
    The data sets are passed (by calling setDataSets() or at instantiation time) as a dictionary::

        datadict={name:(x,y),...}

    where name is the curve name and
    x,y are iterable containers (e.g., lists, tuple, arrays...) of data to be exported

    @TODO: It would be nice if the textedit scrolled to the start ***also for the first set loaded***"""

    # constants
    allInSingleFile = "All sets in a single file (table like)"
    allInMultipleFiles = "All set in multiple files"

    def __init__(self, parent=None, datadict=None, sortedNames=None):
        super(QDataExportDialog, self).__init__(parent)
        self.loadUi()
        self._xIsTime = False

        # connections
        self.exportBT.clicked.connect(self.exportData)
        self.dataSetCB.currentIndexChanged['QString'].connect(self.onDataSetCBChange)

        self.setDataSets(datadict, sortedNames)

    def setDataSets(self, datadict, sortedNames=None):
        """Used to set the sets that are to be offered for exporting. It overwrites previous values.
        """
        if datadict is None:
            return
        if sortedNames is None:
            sortedNames = sorted(self.datadict.keys())
        self.sortedNames = sortedNames
        self.datatime = datetime.now()
        self.datadict = datadict
        self.dataSetCB.clear()
        self.dataSetCB.insertItems(0, sortedNames)
        if len(self.datadict) > 1:
            self.dataSetCB.insertItems(
                0, [self.allInSingleFile, self.allInMultipleFiles])

    def exportData(self):
        if self.dataSetCB.currentText() == self.allInMultipleFiles:
            self.exportAllData()
        else:
            self.exportCurrentData()

    def exportCurrentData(self, set=None, ofile=None, verbose=True, AllowCloseAfter=True):
        """Exports data
        Arguments:
        set: the curve name. If none is passed, it uses the one selected by dataSetCB
        ofile: output file name or file handle. It will prompt if not provided
        verbose: set this to False to disable information popups
        AllowCloseAfter: set this to false if you want to ignore the checkbox in the dialog

        """
        if set is None:
            set = str(self.dataSetCB.currentText())

        if ofile is None:
            if set == self.allInSingleFile:
                name = "all.dat"
            else:
                #**lazy** sanitising of the set to *suggest* it as a filename
                name = set.replace('*', '').replace('/', '_').replace('\\', '_')
                name += ".dat"
            ofile, _ = compat.getSaveFileName(self, 'Export File Name', name,
                                              'All Files (*)')
            if not ofile:
                return False
        try:
            if not isinstance(ofile, file):
                ofile = open(str(ofile), "w")
            if self.dataSetCB.currentText() == self.allInMultipleFiles:
                # 1  file per curve
                text = "# DATASET= %s" % set
                text += "\n# SNAPSHOT_TIME= %s\n" % self.datatime.isoformat('_')
                xdata, ydata = self.datadict[set]
                if self.xIsTime():
                    for x,y in zip(xdata, ydata):
                        t = datetime.fromtimestamp(x)
                        text += "%s\t%r\n" % (t.isoformat('_'), y)
                else:
                    for x,y in zip(xdata, ydata):
                        text+="%r\t%r\n" % (x, y)
                print(str(text), file=ofile)
            else:
                print(str(self.dataTE.toPlainText()), file=ofile)
        except:
            Qt.QMessageBox.warning(self,
                                   "File saving failed",
                                   "Failed to save file '%s'" % str(ofile.name),
                                   Qt.QMessageBox.Ok)
            raise
        finally:
            ofile.close()
        if verbose:
            msg = "Set saved to '%s'" % str(ofile.name)
            Qt.QMessageBox.information(self, "Set exported", msg,
                                       Qt.QMessageBox.Ok)
        if AllowCloseAfter and self.closeAfterCB.isChecked(): 
            self.accept() #closes the ExportData dialog with Accept state
        return True

    def exportAllData(self, preffix=None):
        """Exports all sets using a common preffix and appending 'XXX.dat', where XXX is a number starting at 001
        if preffix is not given, the user is prompted for a directory path"""
        if preffix is None:
            outputdir = Qt.QFileDialog.getExistingDirectory(
                self, 'Export Directory', '')
            if not outputdir:
                return False
            preffix = os.path.join(str(outputdir), "set")
        for i, k in zip(range(len(self.datadict)), self.sortedNames):
            ofile = "%s%03i.dat" % (preffix, i + 1)
            try:
                self.exportCurrentData(
                    set=k, ofile=ofile, verbose=False, AllowCloseAfter=False)
            except:
                return False
        # mend undesired side effect of updateText in the for loop
        self.updateText(self.allInMultipleFiles)
        Qt.QMessageBox.information(self, "All sets exported", "%i set(s) exported to:\n%sXXX.dat" % (
            len(self.datadict), preffix), Qt.QMessageBox.Ok)
        if self.closeAfterCB.isChecked():
            self.accept()  # closes the ExportData dialog with Accept state
        return True

    def onDataSetCBChange(self, key):
        key = str(key)
        self.updateText(key)

    def updateText(self, key=None):
        '''update the text edit that shows the preview of the data'''
        if key is None:
            key = str(self.dataSetCB.currentText())
        if key in (self.allInMultipleFiles, self.allInSingleFile):
            # check that all arrays have the same length and the same xdata and
            # update header section
            header = "# DATASET= "
            body = ""
            previous = None
            for curve_name in self.sortedNames:
                xdata, ydata = self.datadict[curve_name]
                if previous is None:
                    previous = xdata
                    header += ' "abscissa"'
                elif previous != xdata:
                    if (key == self.allInSingleFile):
                        self.dataTE.clear()
                        Qt.QMessageBox.critical(self,
                                    "Unable to display",
                                    "X axes of all sets in the plot must be " +
                                    "exactly the same for saving in a single " +
                                    "file!. Curves will be saved each one in " +
                                    "its own file",
                                    Qt.QMessageBox.Ok)
                        index = self.dataSetCB.findText(self.allInMultipleFiles)
                        self.dataSetCB.setCurrentIndex(index)
                        return
                    else:
                        self.dataTE.clear()
                        self.dataTE.insertPlainText("Unable to display because abscissas are different.\n"
                                                    "Curves will be saved each one in its own file")
                        return
                header += ' , "%s"' % curve_name

            header += "\n# SNAPSHOT_TIME= %s\n" % self.datatime.isoformat('_')
            # if we reached this point x axes are equal, so fill the editor
            # with the data
            for i, x in enumerate(previous):
                if self.xIsTime():
                    t = datetime.fromtimestamp(x)
                    body += "%s" % t.isoformat('_')
                else:
                    body += "%r" % x
                for curve_name in self.sortedNames:
                    xdata, ydata = self.datadict[curve_name]
                    body += ("\t%r" % ydata[i])
                body += "\n"
            # fill text editor
            self.dataTE.clear()
            self.dataTE.insertPlainText(header + body)
            self.dataTE.moveCursor(Qt.QTextCursor.Start)
            if key == self.allInMultipleFiles:
                self.dataTE.setReadOnly(True)
            else:
                self.dataTE.setReadOnly(False)
        else:
            self.dataTE.setReadOnly(False)
            xdata, ydata = self.datadict[key]
            text = '# DATASET= "%s"\n' % key
            text += "# SNAPSHOT_TIME= %s\n" % self.datatime.isoformat('_')
            if self.xIsTime():
                for x, y in zip(xdata, ydata):
                    t = datetime.fromtimestamp(x)
                    text += "%s\t%r\n"%(t.isoformat('_'), y)
            else:
                for x, y in zip(xdata, ydata):
                    text += "%r\t%r\n" %(x, y)
            self.dataTE.clear()
            self.dataTE.insertPlainText(text)
            self.dataTE.moveCursor(Qt.QTextCursor.Start)

    def setXIsTime(self, xIsTime):
        self._xIsTime = xIsTime
        self.updateText()

    def xIsTime(self):
        return self._xIsTime


if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)
    form = QDataExportDialog()
    form.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

##############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""helper methods for taurus sphinx documentation"""

from __future__ import print_function


__expr = ('or',)


def process_type(t, obj_type='class'):
    t = t.strip()
    if not t:
        return ''
    if t in __expr:
        return t
    if t.count(' or '):
        i = t.index(' or ')
        return ' '.join(map(process_type, (t[:i], 'or', t[i + 4:])))
    if not t.count('<') or not t.count('>'):
        ret = ':%s:`~%s`' % (obj_type, t)
        return ret

    # process a container template
    start, stop = t.index('<'), t.index('>')
    main_type = t[:start]
    main_type = process_type(main_type)
    types = t[start + 1:stop].split(',')
    types = ', '.join(map(process_type, types))
    return "%s <%s>" % (main_type, types)


def process_param(line):
    new_lines = []
    try:
        prefix, param, desc = line.split(':', 2)
        p, param_name = param.split()
        desc = desc.strip()
        if desc[0] == '(':
            pos = desc.find(')')
            if pos != -1:
                elem_type = desc[1:pos]
                klass = process_type(elem_type)
                desc = desc[pos + 1:]
                new_lines.append('%s:type %s: %s' %
                                 (prefix, param_name, klass))
        new_lines.append('%s:param %s: %s' % (prefix, param_name, desc))
    except Exception as e:
        print("Taurus sphinx extension: Not able to process param: '%s'" % line)
        print("      Reason:", str(e))
        new_lines.append(line)
    return new_lines


def process_return(line):
    new_lines = []
    try:
        prefix, param, desc = line.split(':', 2)
        desc = desc.strip()
        if desc[0] == '(':
            pos = desc.find(')')
            if pos != -1:
                elem_type = desc[1:pos]
                klass = process_type(elem_type)
                desc = desc[pos + 1:]
                new_lines.append('%s:rtype: %s' % (prefix, klass))
        new_lines.append('%s:return: %s' % (prefix, desc))
    except Exception as e:
        print("TaurusExtension: Not able to process 'return': '%s'" % line)
        print("      Reason:", str(e))
        new_lines.append(line)
    return new_lines


def process_raise(line):
    new_lines = []
    try:
        prefix, param, desc = line.split(':', 2)
        desc = desc.strip()
        klass = ''
        if desc[0] == '(':
            pos = desc.find(')')
            if pos != -1:
                elem_type = desc[1:pos]
                klass = "(" + process_type(elem_type, obj_type='exc') + ")"
                desc = desc[pos + 1:]
        new_lines.append('%s:raise: %s %s' % (prefix, klass, desc))
    except Exception as e:
        print("TaurusExtension: Not able to process 'raise': '%s'" % line)
        print("      Reason:", str(e))
        new_lines.append(line)
    return new_lines


def _is_return(line):
    ret = line.startswith(':return')
    ret |= line.startswith(':returns')
    return ret


def _is_param(line):
    ret = line.startswith(':param')
    ret |= line.startswith(':parameter')
    ret |= line.startswith(':arg')
    ret |= line.startswith(':argument')
    ret |= line.startswith(':key')
    ret |= line.startswith(':keyword')
    return ret


def _is_raise(line):
    ret = line.startswith(':raise')
    ret |= line.startswith(':except')
    return ret


def process_signature(app, what, name, obj, options, lines):
    ret = []
    for nb, line in enumerate(lines):
        line_strip = line.strip()
        if _is_param(line_strip):
            ret.extend(process_param(line))
        elif _is_return(line_strip):
            ret.extend(process_return(line))
        elif _is_raise(line_strip):
            ret.extend(process_raise(line))
        else:
            ret.append(line)

    del lines[:]
    lines.extend(ret)


def setup(app):
    app.connect('autodoc-process-docstring', process_signature)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
nexusWidget.py:
"""

from builtins import str

import numpy
import posixpath
from functools import partial

from taurus.external.qt import Qt, compat
from PyMca5.PyMcaGui.io.hdf5 import HDF5Widget, HDF5Info, HDF5DatasetTable

from taurus.qt.qtgui.container import TaurusWidget


__all__ = ["TaurusNexusBrowser"]


class NeXusInfoWidget(Qt.QTabWidget):

    def __init__(self, parent=None, info=None):
        Qt.QTabWidget.__init__(self, parent)

        self.generalInfoWidget = HDF5GeneralInfoWidget(self)
        self.attributesInfoWidget = HDF5AttributesInfoWidget(self)
        self.addTab(self.generalInfoWidget, 'General')
        self.addTab(self.attributesInfoWidget, 'Attributes')
        self.__previewWidget = None

    def setInfoDict(self, info):
        self.__previewWidget = NeXusPreviewWidgetFactory(info)


class TaurusNeXusBrowser(TaurusWidget):
    ''' A Browser for nexus files with optional preview. Based on PyMCA's HDF5Widget'''

    def __init__(self, *args, **kwargs):
        TaurusWidget.__init__(self, *args)

        fileModel = kwargs.get('fileModel', None)
        if fileModel is None:
            fileModel = HDF5Widget.FileModel()
        self.__fileModel = fileModel

        self.treeWidget = HDF5Widget.HDF5Widget(self.__fileModel)
        self.treeWidget.setSizePolicy(Qt.QSizePolicy(
            Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Expanding))
#        self.infoWidget = HDF5Info.HDF5InfoWidget()
        self.__previewStack = Qt.QStackedWidget()
        self.__currentPreview = None

        # layout
        self.__splitter = Qt.QSplitter()
        self.__splitter.setOrientation(Qt.Qt.Vertical)
        self.__splitter.addWidget(self.treeWidget)
#        self.__splitter.addWidget(self.infoWidget)
        self.__splitter.addWidget(self.__previewStack)
        self.setLayout(Qt.QVBoxLayout())
        self.layout().addWidget(self.__splitter)

        # Actions
        self.setContextMenuPolicy(Qt.Qt.ActionsContextMenu)
        self.openFileAction = Qt.QAction(Qt.QIcon.fromTheme("document-open"),
                                         "Open Data File...", self)
        self.togglePreviewAction = Qt.QAction(Qt.QIcon("actions:view.svg"),
                                              "Show/Hide preview", self)
        self.togglePreviewAction.setCheckable(True)
        self.togglePreviewAction.setChecked(True)
        self.addActions([self.openFileAction, self.togglePreviewAction])

        # Toolbar
        self._toolbar = Qt.QToolBar("NeXus browser toolbar")
        self._toolbar.setIconSize(Qt.QSize(16, 16))
        self._toolbar.setFloatable(False)
        self._toolbar.addActions(
            [self.openFileAction, self.togglePreviewAction])
        self.layout().setMenuBar(self._toolbar)

        # connections
        self.__fileModel.sigFileAppended.connect(self.treeWidget.fileAppended)
        self.treeWidget.sigHDF5WidgetSignal.connect(self.onHDF5WidgetSignal)
        self.openFileAction.triggered.connect(
            partial(self.openFile, fname=None))
        self.togglePreviewAction.toggled.connect(self.__previewStack.setVisible)

        # configuration
        self.registerConfigProperty(
            self.togglePreviewAction.isChecked, self.togglePreviewAction.setChecked, 'showPreview')

    @Qt.pyqtSlot()
    @Qt.pyqtSlot('QString')
    def openFile(self, fname=None):
        if fname is None:
            fname, _ = compat.getOpenFileName(self, 'Choose NeXus File', '')
        if fname:
            self.__nexusFile = self.__fileModel.openFile(fname)

    def onHDF5WidgetSignal(self, ddict):
        self.__previewStack.removeWidget(self.__currentPreview)
        self.__currentPreview = self.neXusPreviewWidgetFactory(ddict)
        self.__previewStack.addWidget(self.__currentPreview)
        self.__previewStack.setCurrentWidget(self.__currentPreview)

    def neXusPreviewWidgetFactory(self, ddict):
        '''returns a widget showing a preview of a node in a NeXus file'''
        #@todo: check if this is the root group and it has a valid nexus data directory, in which case the default plot should be returned
        if ddict['type'] == 'Dataset':
            node = ddict['name']
            data = self.__nexusFile[node]
            if len(data.shape) == 1 and isinstance(data[0], (numpy.floating, numpy.integer, int, float)):
                try:
                    import pyqtgraph as pg
                    w = pg.PlotWidget()
                    w.plot(data)
                except ImportError:
                    w = HDF5DatasetTable.HDF5DatasetTable()
                    w.setDataset(data)
            else:
                w = HDF5DatasetTable.HDF5DatasetTable()
                w.setDataset(data)
        else:
            info = HDF5Info.getInfo(self.__nexusFile, ddict["name"])
            w = HDF5Info.HDF5InfoWidget()
            w.setInfoDict(info)
        return w

    def neXusWidget(self):
        return self.treeWidget

    def findNodeIndex(self, filename, nodename):
        nexus_widget = self.neXusWidget()
        file_model = nexus_widget.model()

        for node in file_model.rootItem.children:
            if node.file.filename == filename:
                file_node = node
                break
        else:
            raise Exception("Could not find file %s" % filename)

        index = file_model.index(file_node.row, 0, Qt.QModelIndex())

        node_parts = nodename.split(posixpath.sep)
        while node_parts:
            name = posixpath.basename(node_parts.pop(0))
            for child in node.children:
                child_name = posixpath.basename(child.name)
                if child_name == name:
                    node = child
                    index = file_model.index(node.row, 0, index)
                    break
            else:
                raise Exception("Could not find node %s in %s" %
                                (name, filename))
        return index

    def setCurrentNode(self, filename, nodename):
        index = self.findNodeIndex(filename, nodename)
        self.setCurrentIndex(index)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.extra_nexus'
        ret['icon'] = "designer:listview.png"
        ret['container'] = False
        ret['group'] = 'Taurus Views'
        return ret


if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    if len(sys.argv) > 1:
        fname = sys.argv[1]
    else:
        fname = None

    app = TaurusApplication(sys.argv)
    w = TaurusNeXusBrowser()
    w.openFile(fname)
    w.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""The core module"""
from __future__ import absolute_import

from . import release as Release
# from .enums import * #note: all the enums from enums.py were moved to
# taurusbasetypes.py
from .taurusbasetypes import *
from .taurusexception import *
from .taurusmodel import *
from .tauruslistener import *
from .taurusdevice import *
from .taurusattribute import *
from .taurusconfiguration import *
from .taurusauthority import *
from .taurusfactory import *
from .taurusmanager import *
from .taurusoperation import *
from .tauruspollingtimer import *
from .taurusvalidator import *

# enable compatibility code with tau V1 if tauv1 package is present
try:
    from .tauv1 import *
except:
    pass

__docformat__ = "restructuredtext"



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Extension of :mod:`guiqwt.plot`
"""
from builtins import next
from builtins import str

import copy

from future.utils import string_types
from guiqwt.plot import ImageDialog, CurveDialog

import taurus.core
from taurus.external.qt import Qt

from taurus.core.util.containers import CaselessList
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_ATTR_MIME_TYPE
from taurus.qt.qtgui.extra_guiqwt.builder import make
from taurus.qt.qtgui.extra_guiqwt.curve import TaurusCurveItem, TaurusTrendParam, TaurusTrendItem


__all__ = ["TaurusCurveDialog", "TaurusTrendDialog", "TaurusImageDialog"]


class TaurusCurveDialog(CurveDialog, TaurusBaseWidget):
    '''A taurus dialog for showing 1D data.
    It behaves as a regular :class:`guiqwt.plot.CurveDialog` but it also offers
    the expected Taurus interface (e.g. setting models, save/apply configs,
    drag&drops,...)

    .. seealso:: :class:`TaurusCurveWidget`
    '''
    _modifiableByUser = True
    modelChanged = Qt.pyqtSignal([], ['QStringList'], ['QString'])

    def __init__(self, parent=None, designMode=False, toolbar=True, **kwargs):
        '''see :class:`guiqwt.plot.CurveDialog` for other valid initialization parameters'''
        CurveDialog.__init__(self, parent=parent, toolbar=toolbar, **kwargs)
        TaurusBaseWidget.__init__(self, 'TaurusCurveDialog')
        self.deprecated(rel='4.1', dep='TaurusCurveDialog', alt='TaurusPlot / taurusplot launcher')
        self.setWindowFlags(Qt.Qt.Widget)
        self._designMode = designMode
        self._modelNames = CaselessList()
        from guiqwt.styles import style_generator
        self.style = style_generator()
        self.setSupportedMimeTypes(
            [TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_ATTR_MIME_TYPE])
        from taurus.qt.qtgui.extra_guiqwt.tools import TaurusCurveChooserTool
        self.add_tool(TaurusCurveChooserTool)
        self.setModifiableByUser(self._modifiableByUser)
        self.setContextMenuPolicy(Qt.Qt.CustomContextMenu)

    def keyPressEvent(self, event):
        if(event.key() == Qt.Qt.Key_Escape):
            event.ignore()
        else:
            ImageDialog.keyPressEvent(self, event)

    def getModelClass(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return taurus.core.taurusattribute.TaurusAttribute

    def _splitModel(self, modelNames):
        '''convert str to list if needed (commas and whitespace are considered as separators)'''
        if isinstance(modelNames, string_types):
            modelNames = str(modelNames).replace(',', ' ')
            modelNames = modelNames.split()
        return modelNames

    @Qt.pyqtSlot('QStringList')
    def setModel(self, modelNames):
        '''Removes current TaurusCurveItems and adds new ones.

        :param modelNames: (sequence<str> or str) the names of the models to be
                           plotted. For convenience, a string is also accepted
                           (instead of a sequence of strings), in which case the
                           string will be internally converted to a sequence by
                           splitting it on whitespace and commas. Each model can
                           optionally be composed of two parts, separated by "|"
                           indicating X and Y components for the curve. If only
                           one part is given, it is used for Y and X is
                           automatically generated as an index.

        .. seealso:: :meth:`addModels`
        '''

        plot = self.get_plot()
        # delete current TaurusCurveItems
        taurusCurveItems = [item for item in plot.get_public_items(
        ) if isinstance(item, TaurusCurveItem)]
        plot.del_items(taurusCurveItems)
        self._modelNames = CaselessList()
        # add new TaurusCurveItems
        self.addModels(modelNames)

    def addModels(self, modelNames):
        '''Creates TaurusCurveItems (one for each model in modelNames) and attaches
        them to the plot.

        .. note:: you can also add curves using :meth:`add_items`. :meth:`addModels`
                  is only a more Taurus-oriented interface. :meth:`add_items`
                  gives you more control.

        :param modelNames: (sequence<str> or str) the names of the models to be
                           plotted. For convenience, string is also accepted
                           (instead of a sequence of strings), in which case the
                           string will be internally converted to a sequence by
                           splitting it on whitespace and commas. Each model can
                           optionally be composed of two parts, separated by "|"
                           indicating X and Y components for the curve. If only
                           one part is given, it is used for Y and X is
                           automatically generated as an index.

        .. seealso:: :meth:`add_item`
        '''
        plot = self.get_plot()

        # pre-process the model names
        modelNames = self._splitModel(modelNames)
        self._modelNames.extend([str(n) for n in modelNames])
        if self._designMode:
            return
        # create and attach new TaurusCurveItems
        for m in modelNames:
            # split model into x and y components
            mx_my = m.split('|')
            n = len(mx_my)
            if n == 1:
                mx, my = None, mx_my[0]
            elif n == 2:
                mx, my = mx_my
            else:
                self.warning('Invalid model "%s" (Skipping)' % mx_my)
            # cycle styles
            style = next(self.style)
            color = style[0]
            linestyle = style[1:]
            # add the item
            item = make.curve(mx, my, color=color,
                              linestyle=linestyle, linewidth=2)
            item.set_readonly(not self.isModifiableByUser())
            plot.add_item(item)
        self.modelChanged.emit()

    def getModel(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        return self._modelNames

    def setModifiableByUser(self, modifiable):
        """reimplemented from :class:`TaurusBaseWidget`"""
        from taurus.qt.qtgui.extra_guiqwt.tools import TaurusCurveChooserTool
        self.get_tool(TaurusCurveChooserTool).action.setEnabled(modifiable)
        self.get_plot().set_items_readonly(not modifiable)
        TaurusBaseWidget.setModifiableByUser(self, modifiable)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """reimplemented from :class:`TaurusBaseWidget`"""
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.extra_guiqwt'
        ret['group'] = 'Taurus Display'
        ret['icon'] = 'designer:qwtplot.png'
        return ret

    model = Qt.pyqtProperty("QStringList", getModel,
                            setModel, TaurusBaseWidget.resetModel)
    modifiableByUser = Qt.pyqtProperty(
        "bool", TaurusBaseWidget.isModifiableByUser, setModifiableByUser, TaurusBaseWidget.resetModifiableByUser)


class TaurusTrendDialog(CurveDialog, TaurusBaseWidget):
    '''A taurus widget for showing trends of scalar data.
    It is an specialization of :class:`guiqwt.plot.CurveWidget`, for displaying
    trends and offering the expected Taurus interface (e.g. setting models,
    save/apply configs, drag&drops,...)

    .. seealso:: :class:`TaurusTrendDialog`
    '''
    _modifiableByUser = True

    modelChanged = Qt.pyqtSignal([], ['QStringList'], ['QString'])

    def __init__(self, parent=None, designMode=False, taurusparam=None, toolbar=True, **kwargs):
        '''see :class:`guiqwt.plot.CurveDialog` for other valid initialization parameters'''
        CurveDialog.__init__(self, parent=parent, toolbar=toolbar, **kwargs)
        TaurusBaseWidget.__init__(self, 'TaurusTrendDialog')
        self.deprecated(rel='4.1', dep='TaurusTrendDialog', alt='TaurusTrend / taurustrend launcher')
        self.setWindowFlags(Qt.Qt.Widget)
        self._designMode = designMode
        self._modelNames = CaselessList()
        from guiqwt.styles import style_generator
        self.style = style_generator()
        self.setSupportedMimeTypes(
            [TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_ATTR_MIME_TYPE])
        from taurus.qt.qtgui.extra_guiqwt.tools import TaurusModelChooserTool, AutoScrollTool
        self.add_tool(TaurusModelChooserTool, singleModel=False)
        self.add_tool(AutoScrollTool)
        self.setModifiableByUser(self._modifiableByUser)
        if taurusparam is None:
            taurusparam = TaurusTrendParam()
        self.defaultTaurusparam = taurusparam
        self.setContextMenuPolicy(Qt.Qt.CustomContextMenu)
        self.registerConfigDelegate(self.get_tool(AutoScrollTool))

    def keyPressEvent(self, event):
        if(event.key() == Qt.Qt.Key_Escape):
            event.ignore()
        else:
            ImageDialog.keyPressEvent(self, event)

    def getModelClass(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return taurus.core.taurusattribute.TaurusAttribute

    def getTaurusTrendItems(self):
        return [item for item in self.get_plot().get_public_items() if isinstance(item, TaurusTrendItem)]

    def _splitModel(self, modelNames):
        '''convert str to list if needed (commas and whitespace are considered as separators)'''
        if isinstance(modelNames, string_types):
            modelNames = str(modelNames).replace(',', ' ')
            modelNames = modelNames.split()
        return modelNames

    @Qt.pyqtSlot('QStringList')
    def setModel(self, modelNames):
        '''Removes current TaurusCurveItems and adds new ones.

        :param modelNames: (sequence<str> or str) the names of the models to be
                           plotted. For convenience, a string is also accepted
                           (instead of a sequence of strings), in which case the
                           string will be internally converted to a sequence by
                           splitting it on whitespace and commas.

        .. seealso:: :meth:`addModels`
        '''

        plot = self.get_plot()
        # delete current TaurusCurveItems
        taurusTrendItems = self.getTaurusTrendItems()
        plot.del_items(taurusTrendItems)
        self._modelNames = CaselessList()
        # add new TaurusCurveItems
        self.addModels(modelNames)

    def addModels(self, modelNames):
        '''Creates TaurusCurveItems (one for each model in modelNames) and attaches
        them to the plot.

        .. note:: you can also add curves using :meth:`add_items`. :meth:`addModels`
                  is only a more Taurus-oriented interface. :meth:`add_items`
                  gives you more control.

        :param modelNames: (sequence<str> or str) the names of the models to be
                           plotted. For convenience, a string is also accepted
                           (instead of a sequence of strings), in which case the
                           string will be internally converted to a sequence by
                           splitting it on whitespace and commas.

        .. seealso:: :meth:`add_item`
        '''
        plot = self.get_plot()

        # pre-process the model names
        modelNames = self._splitModel(modelNames)
        self._modelNames.extend([str(n) for n in modelNames])
        if self._designMode:
            return
        # create and attach new TaurusCurveItems
        for m in modelNames:
            # cycle styles
            style = next(self.style)
            # add the item
            item = make.ttrend(m, color=style[0], linestyle=style[
                               1:], linewidth=2, taurusparam=copy.deepcopy(self.defaultTaurusparam))
            item.set_readonly(not self.isModifiableByUser())
            plot.add_item(item)
            item.update_params()

        self.setStackMode(self.defaultTaurusparam.stackMode)
        self.modelChanged.emit()

    def getModel(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        return self._modelNames

    def setUseArchiving(self, enable):
        '''enables/disables looking up in the archiver for data stored before
        the Trend was started

        :param enable: (bool) if True, archiving values will be used if available
        '''
        if not self.defaultTaurusparam.stackMode == 'datetime':
            self.info('ignoring setUseArchiving. Reason: not in X time scale')
        self.defaultTaurusparam.useArchiving = enable

    def getUseArchiving(self):
        '''whether TaurusTrend is looking for data in the archiver when needed

        :return: (bool)

        .. seealso:: :meth:`setUseArchiving`
        '''
        return self.defaultTaurusparam.useArchiving

    def resetUseArchiving(self):
        '''Same as setUseArchiving(False)'''
        self.setUseArchiving(False)

    def setMaxDataBufferSize(self, maxSize):
        '''sets the maximum number of events that will be stacked

        :param maxSize: (int) the maximum limit

        .. seealso:: :class:`TaurusTrendSet`
        '''
        for item in self.getTaurusTrendItems():
            item.setBufferSize(maxSize)

        self.defaultTaurusparam.maxBufferSize = maxSize

    def getMaxDataBufferSize(self):
        '''returns the maximum number of events that can be plotted in the trend

        :return: (int)
        '''
        return self.defaultTaurusparam.maxBufferSize

    def resetMaxDataBufferSize(self):
        '''Same as setMaxDataBufferSize(16384)'''
        self.setMaxDataBufferSize(16384)

    def setStackMode(self, mode):
        '''set the type of stack to be used. This determines how X values are
        interpreted:

            - as timestamps ('datetime')
            - as time deltas ('timedelta')
            - as event numbers ('event')

        :param mode:(one of 'datetime', 'timedelta' or 'event')
        '''
        from taurus.qt.qtgui.extra_guiqwt.tools import TimeAxisTool
        mode = str(mode)
        if mode == 'datetime':
            timetool = self.get_tool(TimeAxisTool)
            if timetool is None:
                self.add_tool(TimeAxisTool)
                timetool = self.get_tool(TimeAxisTool)
            timetool.set_scale_y_t(True)
        elif mode == 'deltatime':
            from taurus.qt.qtgui.extra_guiqwt.scales import DeltaTimeScaleEngine
            plot = self.get_plot()
            DeltaTimeScaleEngine.enableInAxis(plot, plot.xBottom, rotation=-45)
        elif mode == 'event':
            plot = self.get_plot()
            scaleEngine = plot.axisScaleEngine(plot.xBottom)
            if hasattr(scaleEngine, 'disableInAxis'):
                scaleEngine.disableInAxis(plot, plot.xBottom)
        else:
            self.error('Unknown stack mode "%s"' % repr(mode))
            return

        self.defaultTaurusparam.stackMode = mode

        for item in self.getTaurusTrendItems():
            item.taurusparam.stackMode = mode

    def getStackMode(self):
        return self.defaultTaurusparam.stackMode

    def resetStackMode(self):
        self.setStackMode('datetime')

    def setModifiableByUser(self, modifiable):
        """reimplemented from :class:`TaurusBaseWidget`"""
        from taurus.qt.qtgui.extra_guiqwt.tools import TaurusModelChooserTool
        self.get_tool(TaurusModelChooserTool).action.setEnabled(modifiable)
        self.get_plot().set_items_readonly(not modifiable)
        TaurusBaseWidget.setModifiableByUser(self, modifiable)

    def getDropEventCallback(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        return self.addModels

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """reimplemented from :class:`TaurusBaseWidget`"""
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.extra_guiqwt'
        ret['group'] = 'Taurus Display'
        ret['icon'] = 'designer:qwtplot.png'
        return ret

    model = Qt.pyqtProperty("QStringList", getModel,
                            setModel, TaurusBaseWidget.resetModel)
    useArchiving = Qt.pyqtProperty(
        "bool", getUseArchiving, setUseArchiving, resetUseArchiving)
    maxDataBufferSize = Qt.pyqtProperty(
        "int", getMaxDataBufferSize, setMaxDataBufferSize, resetMaxDataBufferSize)
    stackMode = Qt.pyqtProperty(
        "QString", getStackMode, setStackMode, resetStackMode)
    modifiableByUser = Qt.pyqtProperty(
        "bool", TaurusBaseWidget.isModifiableByUser, setModifiableByUser, TaurusBaseWidget.resetModifiableByUser)


class TaurusImageDialog(ImageDialog, TaurusBaseWidget):
    '''A taurus dialog for showing 2D data.
    It behaves as a regular :class:`guiqwt.plot.ImageDialog` but it also offers
    the expected Taurus interface (e.g. setting models, save/apply configs,
    drag&drops,...)

    .. seealso:: :class:`TaurusImageWidget`
    '''
    _rgbmode = False

    def __init__(self, parent=None, designMode=False, toolbar=True, **kwargs):
        '''see :class:`guiqwt.plot.ImageDialog` for other valid initialization parameters'''
        ImageDialog.__init__(self, parent=parent, toolbar=toolbar, **kwargs)
        TaurusBaseWidget.__init__(self, 'TaurusImageDialog')
        self.setWindowFlags(Qt.Qt.Widget)
        self.imgItem = None
        from taurus.qt.qtgui.extra_guiqwt.tools import TaurusModelChooserTool
        self.add_tool(TaurusModelChooserTool, singleModel=True)
        self.setModifiableByUser(True)
        self.setContextMenuPolicy(Qt.Qt.CustomContextMenu)

    def keyPressEvent(self, event):
        if(event.key() == Qt.Qt.Key_Escape):
            event.ignore()
        else:
            ImageDialog.keyPressEvent(self, event)

    def setRGBmode(self, enable):
        self._rgbmode = enable
        # enable the zaxis only when not in rgb mode
        plot = self.get_plot()
        zaxis = plot.colormap_axis
        plot.enableAxis(zaxis, not enable)

    def getRGBmode(self):
        return self._rgbmode

    def resetRGBmode(self):
        self.setRGBmode(self.__class__._rgbmode)

    def getModelClass(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return taurus.core.taurusattribute.TaurusAttribute

    @Qt.pyqtSlot(str)
    def setModel(self, model):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        if self.getUseParentModel():
            # @fixme: This assumes Tango models.
            model = "/".join((str(self.getParentModelName()), str(model)))
        plot = self.get_plot()
        if self.imgItem is not None:
            try:
                plot.del_item(self.imgItem)
            except:
                self.info("Unable to delete item from plot")
        if not model:
            self.imgItem = None
            return
        if self.rgbmode:
            self.imgItem = make.rgbimage(taurusmodel=model)
        else:
            self.imgItem = make.image(taurusmodel=model)
        plot.add_item(self.imgItem)
        self.imgItem.set_readonly(not self.isModifiableByUser())
        # IMPORTANT: connect the cross section plots to the taurusimage so that
        # they are updated when the taurus data changes
        self.imgItem.dataChanged.connect(self.update_cross_sections)

    def getModel(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        if self.imgItem is None:
            return ''
        else:
            return self.imgItem.getModel()

    def setModifiableByUser(self, modifiable):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        from taurus.qt.qtgui.extra_guiqwt.tools import TaurusModelChooserTool
        self.get_tool(TaurusModelChooserTool).action.setEnabled(modifiable)
        self.get_plot().set_items_readonly(not modifiable)
        TaurusBaseWidget.setModifiableByUser(self, modifiable)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """reimplemented from :class:`TaurusBaseWidget`"""
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.extra_guiqwt'
        ret['group'] = 'Taurus Display'
        ret['icon'] = 'designer:qwtplot.png'
        return ret

    model = Qt.pyqtProperty("QString", getModel, setModel,
                            TaurusBaseWidget.resetModel)
    rgbmode = Qt.pyqtProperty("bool", getRGBmode, setRGBmode, resetRGBmode)
    useParentModel = Qt.pyqtProperty("bool", TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel,
                                     TaurusBaseWidget.resetUseParentModel)
    modifiableByUser = Qt.pyqtProperty("bool", TaurusBaseWidget.isModifiableByUser,
                                       setModifiableByUser,
                                       TaurusBaseWidget.resetModifiableByUser)


def taurusCurveDlgMain():
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.qt.qtgui.extra_guiqwt.tools import TimeAxisTool
    import taurus.core.util.argparse
    import sys

    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.set_usage("%prog [options] [<model1> [<model2>] ...]")
    parser.set_description("a taurus application for plotting 1D data sets")
    parser.add_option("--demo", action="store_true", dest="demo",
                      default=False, help="show a demo of the widget")
    parser.add_option("--window-name", dest="window_name",
                      default="Taurus Curve Dialog", help="Name of the window")
    app = TaurusApplication(
        cmd_line_parser=parser, app_name="Taurus Curve Dialog", app_version=taurus.Release.version)
    args = app.get_command_line_args()
    options = app.get_command_line_options()

    # check & process options
    if options.demo:
        args.append('eval:rand(128)')

    w = TaurusCurveDialog(edit=False, wintitle=options.window_name)

    w.add_tool(TimeAxisTool)

    # set model
    if len(args) > 0:
        w.setModel(args)
    else:
        parser.print_help(sys.stderr)
        sys.exit(1)

    w.show()
    sys.exit(app.exec_())


def taurusTrendDlgMain():
    from taurus.qt.qtgui.application import TaurusApplication
    import taurus.core
    import sys

    # prepare options
    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.set_usage("%prog [options] <model>")
    parser.set_description(
        'a Taurus application for plotting trends of scalars')
    parser.add_option("-x", "--x-axis-mode", dest="x_axis_mode", default='t', metavar="t|d|e",
                      help="interpret X values as timestamps (t), time deltas (d) or event numbers (e). Accepted values: t|d|e")
    parser.add_option("-b", "--buffer", dest="max_buffer_size", default='10000',
                      help="maximum number of values to be plotted (when reached, the oldest values will be discarded)")
    parser.add_option("-a", "--use-archiving",
                      action="store_true", dest="use_archiving", default=False)
    parser.add_option("--demo", action="store_true", dest="demo",
                      default=False, help="show a demo of the widget")
    parser.add_option("--window-name", dest="window_name",
                      default="Taurus Trend", help="Name of the window")
    app = TaurusApplication(
        cmd_line_parser=parser, app_name="Taurus Trend", app_version=taurus.Release.version)
    args = app.get_command_line_args()
    options = app.get_command_line_options()

    # check & process options
    stackModeMap = dict(t='datetime', d='deltatime', e='event')
    if options.x_axis_mode.lower() not in stackModeMap:
        parser.print_help(sys.stderr)
        sys.exit(1)

    stackMode = stackModeMap[options.x_axis_mode.lower()]

    if options.use_archiving:
        raise NotImplementedError('Archiving support is not yet implemented')

    if options.demo:
        args.append('eval:rand()')

    taurusparam = TaurusTrendParam()
    taurusparam.stackMode = stackMode
    taurusparam.maxBufferSize = int(options.max_buffer_size)
    taurusparam.useArchiving = options.use_archiving

    w = TaurusTrendDialog(wintitle=options.window_name,
                          taurusparam=taurusparam)

    # set model
    if len(args) > 0:
        w.setModel(args)
    else:
        parser.print_help(sys.stderr)
        sys.exit(1)

    w.show()
    sys.exit(app.exec_())


def taurusImageDlgMain():
    from taurus.qt.qtgui.application import TaurusApplication
    import taurus.core
    import sys

    # prepare options
    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.set_usage("%prog [options] <model>")
    parser.set_description(
        'a Taurus application for plotting Image Attributes')
    parser.add_option("--demo", action="store_true", dest="demo",
                      default=False, help="show a demo of the widget")
    parser.add_option("--rgb", action="store_true", dest="rgb_mode",
                      default=False, help="assume image is RGB")
    parser.add_option("--window-name", dest="window_name",
                      default="Taurus Image", help="Name of the window")
    app = TaurusApplication(
        cmd_line_parser=parser, app_name="Taurus Image Dialog", app_version=taurus.Release.version)
    args = app.get_command_line_args()
    options = app.get_command_line_options()

    # check & process options
    if options.demo:
        if options.rgb_mode:
            args.append('eval:randint(0,256,(10,20,3))')
        else:
            args.append('eval:rand(256,128)')
    w = TaurusImageDialog(wintitle=options.window_name)

    w.setRGBmode(options.rgb_mode)

    # set model
    if len(args) == 1:
        w.setModel(args[0])
    else:
        parser.print_help(sys.stderr)
        sys.exit(1)

    w.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    taurusCurveDlgMain()
    # taurusTrendDlgMain()
#    taurusImageDlgMain()



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

from builtins import object
from taurus.core import TaurusElementType
from taurus.core.taurusvalidator import (TaurusAttributeNameValidator,
                                         TaurusDeviceNameValidator,
                                         TaurusAuthorityNameValidator)
from taurus.core.taurushelper import getSchemeFromName, Factory

__all__ = ['ResDeviceNameValidator',
           'ResAttributeNameValidator']

# Pattern for python variables
PY_VAR = r'(?<![\.a-zA-Z0-9_])[a-zA-Z_][a-zA-Z0-9_]*'


class _ResNameValidator(object):
    """
    Abstract class for all Res validators. Res Name validators should inherit
    first from it and then from the TaurusXXXNameValidator to give its methods
    highest precedence according to the python MRO
    """
    scheme = 'res'
    authority = '(?!)'
    path = r'(?P<_resname>%s)' % PY_VAR
    query = '(?!)'
    fragment = '(?!)'

    pattern = r'^(?P<scheme>%(scheme)s):' + \
              r'((?P<authority>%(authority)s)(?=/))?' + \
              r'(?P<path>%(path)s)' + \
              r'(\?(?P<query>%(query)s))?' + \
              r'(#(?P<fragment>%(fragment)s))?$'

    _elementType = None

    def _getValueValidator(self, name):
        """Return the name validator for the mapped model name scheme

        :param name: (str) resource name
        :return: A specific TaurusValidator
        """
        scheme = getSchemeFromName(name)
        f = Factory(scheme)
        if self._elementType == TaurusElementType.Attribute:
            return f.getAttributeNameValidator()
        elif self._elementType == TaurusElementType.Device:
            return f.getDeviceNameValidator()
        elif self._elementType == TaurusElementType.Authority:
            return f.getAuthorityNameValidator()
        else:
            msg = '_elementType must be one of (Authority, Device, Attribute)'
            raise Exception(msg)

    def _getKey(self, name):

        if self._elementType == TaurusElementType.Attribute:
            g = TaurusAttributeNameValidator.getUriGroups(self, name)
        elif self._elementType == TaurusElementType.Device:
            g = TaurusDeviceNameValidator.getUriGroups(self, name)
        elif self._elementType == TaurusElementType.Authority:
            g = TaurusAuthorityNameValidator.getUriGroups(self, name)
        else:
            msg = '_elementType must be one of (Authority, Device, Attribute)'
            raise Exception(msg)
        if g is None:
            return None
        else:
            return g['_resname']

    def _getValue(self, key):
        return Factory('res').getValue(key)

    def isValid(self, name, *args, **kwargs):
        """Checks validity of the key (resname) and also of the value
        (model name)"""
        key = self._getKey(name)
        if key is None:
            return False
        value = self._getValue(key)
        if value is None:
            return False
        v = self._getValueValidator(value)
        return v.isValid(value, *args, **kwargs)

    def getUriGroups(self, name, *args, **kwargs):
        """Returns the groups for the value extended with '_resname'"""
        key = self._getKey(name)
        if key is None:
            return None
        value = self._getValue(key)
        if value is None:
            return None
        v = self._getValueValidator(value)
        ret = v.getUriGroups(value, *args, **kwargs)
        if ret is None:
            return None
        ret['_resname'] = key
        return ret

    def getNames(self, name, *args, **kwargs):
        """Returns the names of the value"""
        key = self._getKey(name)
        if key is None:
            return None
        value = self._getValue(key)
        if value is None:
            return None
        return self._getValueValidator(value).getNames(value, *args, **kwargs)


class ResourceAuthorityNameValidator(_ResNameValidator,
                                     TaurusAuthorityNameValidator):
    """Validator for res authority names. Apart from the named related to the
    mapped model value, the following named groups are created:

     - _resname: resource name (aka key)
    """
    # authority = r'(?P<_resname>%s)' % PY_VAR
    # path = '(?!)'
    _elementType = TaurusElementType.Authority


class ResourceDeviceNameValidator(_ResNameValidator,
                                  TaurusDeviceNameValidator):
    """Validator for res device names. Apart from the named related to the
    mapped model value, the following named groups are created:

     - _resname: resource name (aka key)
    """
    _elementType = TaurusElementType.Device


class ResourceAttributeNameValidator(_ResNameValidator,
                                     TaurusAttributeNameValidator):
    """Validator for res attribute names. Apart from the named related to the
    mapped model value, the following named groups are created:

     - _resname: resource name (aka key)
    """
    _elementType = TaurusElementType.Attribute
